#ifndef POLYMAKE_POLYTROPES_GRAPH_ALL_SOLUTIONS_H
#define POLYMAKE_POLYTROPES_GRAPH_ALL_SOLUTIONS_H

#include "polymake/graph/DijkstraShortestPathWithIntervalWeights.h"
#include "polymake/graph/DijkstraShortestPath.h"
#include "polymake/SparseMatrix.h"
#include "polymake/Vector.h"
#include "polymake/Set.h"
#include <deque>

namespace polymake {
namespace graph {

template <typename Weight>
class InitialSolution
   : public DijkstraShortestPathWithIntervalWeights<Weight> {
   using base_layer = DijkstraShortestPathWithIntervalWeights<Weight>;
public:
   template <typename FullLabel = void>
   struct Label : public base_layer::template Label<typename mprefer1st<FullLabel, Label<FullLabel>>::type> {
      using base_t  = typename base_layer::template Label<typename mprefer1st<FullLabel, Label<FullLabel>>::type>;

      template <typename... CostArgs>
      Label(const Int node_arg, CostArgs&&... cost_args)
         : base_t(node_arg, std::forward<CostArgs>(cost_args)...)
         , edge_id(-1) {}

      Int edge_id;
   };

   using typename base_layer::graph_t;
   using typename base_layer::edge_weights_map;

   template <typename Top>
   class Data : public base_layer::template Data<Top> {
      using base_t = typename base_layer::template Data<Top>;
   public:
      using typename base_t::label_t;

      using shortest_path_tree_t = EdgeMap<Directed, bool>;

      class DecisionNodeComparator
         : public operations::cmp {
      public:
         explicit DecisionNodeComparator(const edge_weights_map* edge_weights_arg = nullptr)
            : edge_weights(edge_weights_arg) {}

         // Decision node N1 must be reported as "less than" N2 if the path from N2 to the target node passes through N1.
         // If N1 and N2 are independent, that is, no one is reachable from another one in the shortest path tree,
         // the result may be both, less or greater.  Therefore we try to avoid the expensive search in the tree as far as possible.
         pm::cmp_value operator() (const label_t* l1, const label_t* l2) const
         {
            if (l1 == l2)
               return pm::cmp_eq;
            const pm::cmp_value cmp_cost = operations::cmp::operator()(l1->min_cost, l2->min_cost);
            if (cmp_cost != pm::cmp_eq)
               // the minimal costs are different: node with the smaller cost can't be a descendant of another one
               return cmp_cost;
            const Int cmp_vars = incl(l1->vars, l2->vars);
            if (cmp_vars != 2) {
               if (cmp_vars != 0)
                  // node with a smaller sum of variables in the cost can't be a descendant of another one
                  return pm::cmp_value(cmp_vars);

               // Here we still have to cater for an exotic situation, where the nodes are connected with a path of zero cost
               if (find_on_zero_cost_path(l1->node, l2))
                  return pm::cmp_lt;
               if (find_on_zero_cost_path(l2->node, l1))
                  return pm::cmp_gt;
            }
            // the costs are equal but the nodes are independent; simply compare the node indexes for the sake of determinism
            return operations::cmp::operator()(l1->node, l2->node);
         }
      private:
         bool find_on_zero_cost_path(const Int node, const label_t* label) const
         {
            while ((*edge_weights)[label->edge_id].is_zero() && (label = label->predecessor)) {
               if (label->node == node) return true;
            }
            return false;
         }

         const edge_weights_map* edge_weights;
      };

      //! decision nodes are sorted in increasing order of minimal cost,
      //! so that all decision nodes reachable from a given decision node through the shortest path tree
      //! are guaranteed to occur in the list after that one
      using decision_nodes_list = Set<label_t*, DecisionNodeComparator>;

      Data(const graph_t& G_arg, const edge_weights_map& edge_weights_arg, const Int n_vars_arg)
         : base_t(G_arg, edge_weights_arg, n_vars_arg)
         , shortest_path_tree(G_arg, false)
         , decision_nodes(DecisionNodeComparator(&this->edge_weights))
         , infinite_weight(std::numeric_limits<Weight>::infinity()) {}

      void clear()
      {
         base_t::clear();
         fill_range(entire_range(shortest_path_tree), false);
         decision_nodes.clear();
      }

      shortest_path_tree_t shortest_path_tree;
      decision_nodes_list decision_nodes;
      std::vector<const label_t*> updated_decision_nodes;
      const Weight infinite_weight;
   };

   // provides the initial solution, where all variables have minimal weights
   template <typename Top>
   class Algo : public base_layer::template Algo<Top> {
      using base_t = typename base_layer::template Algo<Top>;
   public:
      using typename base_t::label_t;
      using base_layer::template Algo<Top>::Algo;

      void solve(Int target_node)
      {
         base_t::solve(target_node, [](const label_t&) { return false; }, true);
      }

      bool process_popped(label_t* const pred_label, bool backward) const
      {
         if (base_t::process_popped(pred_label, backward)) {
            if (pred_label->edge_id >= 0)
               this->data.shortest_path_tree[pred_label->edge_id] = true;
            if (pred_label->sibling)
               // by the virtue of Dijkstra's algorithm, popped labels are processed in the increasing order of minimal cost,
               // thus it's save to use push_back here
               this->data.decision_nodes.push_back(pred_label);
            return true;
         }
         return false;
      }

      void add_sibling(label_t* cur_label, label_t* new_label, label_t* pred_label) const
      {
         if (cur_label->heap_pos < 0 && !cur_label->sibling) {
            // the label on an already settled node gets its first sibling
            this->data.decision_nodes.insert(cur_label);
         }
         base_t::add_sibling(cur_label, new_label, pred_label);
      }

      void update_siblings(label_t* cur_label) const
      {
         const bool settled_with_siblings = cur_label->heap_pos < 0 && cur_label->sibling != nullptr;
         base_t::update_siblings(cur_label);
         if (settled_with_siblings && !cur_label->sibling)
            // a settled label lost all its siblings
            this->data.decision_nodes.erase(cur_label);
      }

      label_t* construct_label(const Int node) const
      {
         label_t* label = base_t::construct_label(node);
         label->edge_id = -1;
         return label;
      }

      label_t* construct_label(const label_t* pred_label, Int cur_node, Int edge_id) const
      {
         label_t* label = base_t::construct_label(pred_label, cur_node, edge_id);
         label->edge_id = edge_id;
         return label;
      }
   };
};

template <typename Weight>
class AllSolutions
   : protected DijkstraShortestPath<InitialSolution<Weight>> {
   using init_layer = DijkstraShortestPath<InitialSolution<Weight>>;
   using typename init_layer::data_t;
   using typename init_layer::algo_t;
   using label_t = typename data_t::label_t;

public:
   using shortest_path_tree_t = typename data_t::shortest_path_tree_t;
   using typename init_layer::edge_weights_map;

   using DijkstraShortestPath<InitialSolution<Weight>>::DijkstraShortestPath;

   //! set the target node, generate the first solution: all variables at lower limits
   void start(Int target_node)
   {
      if (!this->data.decision_nodes.empty())
         this->data.clear();
      algo_t algo(this->data);
      algo.solve(target_node);
      this->data.updated_decision_nodes.reserve(this->data.decision_nodes.size());
   }

   //! generate the next solution
   //! @return true when the next solution exists
   bool next();

   //! get the shortest path tree of the current solution
   //! it consists of all source graph edges mapped to 'true'
   const shortest_path_tree_t& get_shortest_path_tree() const
   {
      return this->data.shortest_path_tree;
   }

   //! get the total cost of the path from the given node to the target node in the current solution
   //! @return sparse vector (W | n_0 n_1 ...) encoding the cost W + n_0 * var_0 + n_1 * var_1 + ...
   //!         W = inf when the target node is not reachable from the given node
   SparseVector<Weight> get_cost(Int n) const
   {
      if (const label_t* label = this->data.labels_on_node[n])
         return total_cost(label);

      return SparseVector<Weight>(this->data.n_vars + 1, { { 1, this->data.infinite_weight } });
   }

   //! get the inequalities for variables which are held for the current solution
   //! @return sparse matrix where each row (W n_0 n_1 ...) encodes an inequality
   //!         W + n_0 * var_0 + n_1 * var_1 ... >= 0
   SparseMatrix<Weight> get_inequalities() const
   {
      RestrictedSparseMatrix<Weight, pm::sparse2d::only_cols> ineqs(0, this->data.n_vars + 1);
      for (const label_t* label : this->data.decision_nodes) {
         const Int node = label->node;
         const label_t* active_label = this->data.labels_on_node[node];
         const auto active_cost = total_cost(active_label);
         do {
            if (label != active_label && label->min_cost < active_label->max_cost) {
               SparseVector<Weight> ineq = total_cost(label) - active_cost;
               // filter out trivial inequalities
               if (!find_in_range_if(entire(ineq), operations::negative()).at_end())
                  ineqs /= std::move(ineq);
            }
         } while ((label = label->sibling));
      }
      return SparseMatrix<Weight>(std::move(ineqs));
   }

private:
   //! check whether the sibling can be chosen as an active label in a new solution
   bool can_be_active(const label_t* label) const;

   bool is_better_than_siblings(const label_t* label) const;

   //! update costs, shortest path flags and predecessor links in the subtree rooted at the node owning the active labels
   void update_solution(label_t* old_active_label, label_t* new_active_label, bool collect_decision_points);

   //! check whether decision nodes affected by the most recent sibling swap are still in feasible state
   bool check_updated_decision_nodes() const;

   auto total_cost(const label_t* label) const
   {
      return label->min_cost | convert_to<Weight>(label->vars);
   }
};

template <typename Weight>
bool AllSolutions<Weight>::next()
{
   for (auto decision_it = entire(this->data.decision_nodes);  !decision_it.at_end(); ++decision_it) {
      label_t* best_label = *decision_it;
      const Int node = best_label->node;
      label_t* old_active_label = this->data.labels_on_node[node];
      label_t* new_active_label;
      while ((new_active_label = this->data.rotate_siblings(node)) != best_label) {
         if (can_be_active(new_active_label)) {
            update_solution(old_active_label, new_active_label, true);
            if (check_updated_decision_nodes())
               return true;
            old_active_label = new_active_label;
         }
      }
      // iterated through all siblings: restore the solution for the best label and proceed with the next decision node
      if (old_active_label != best_label)
         update_solution(old_active_label, best_label, false);
   }
   return false;
}

template <typename Weight>
bool AllSolutions<Weight>::can_be_active(const label_t* label) const
{
   const pm::cmp_value cmp_with_upper_bound = operations::cmp()(label->min_cost, label->min_max_cost);
   if (cmp_with_upper_bound == pm::cmp_lt)
      return is_better_than_siblings(label);

   // If the intersection is empty or just a point at the label's lower bound, it can't become better than the best label
   if (cmp_with_upper_bound == pm::cmp_gt || !label->vars.empty())
      return false;

   // The label starts a path of constant weight; it can be chosen if any other label can be worse, that is, the upper bound of the node is defined by this label
   // This routine is always called with the current active label
   const label_t* sibling = label->sibling;
   do {
      if (sibling->max_cost == label->min_max_cost)
         return false;
   } while ((sibling = sibling->sibling));

   return true;
}

template <typename Weight>
bool AllSolutions<Weight>::is_better_than_siblings(const label_t* label) const
{
   const label_t* sibling = label->sibling;

   do {
      if (label->is_worse_and_redundant_by_vars(sibling))
         return false;
   } while ((sibling = sibling->sibling));

   return true;
}

template <typename Weight>
void AllSolutions<Weight>::update_solution(label_t* old_active_label, label_t* new_active_label, const bool collect_decision_points)
{
   this->data.shortest_path_tree[old_active_label->edge_id] = false;
   this->data.shortest_path_tree[new_active_label->edge_id] = true;
   const Weight weight_diff = new_active_label->min_cost - old_active_label->min_cost;
   const SparseVector<Int> vars_diff = new_active_label->vars - old_active_label->vars;

   // update the costs of all labels reachable through the shortest path subtree,
   // flip the predecessor links

   std::deque<label_t*> update_queue;
   for (auto in_edge_it = entire(this->data.G.in_edges(old_active_label->node));  !in_edge_it.at_end();  ++in_edge_it) {
      const Int node_from = in_edge_it.from_node();
      for (label_t* label = this->data.labels_on_node[node_from]; label; label = label->sibling) {
         if (label->predecessor == old_active_label) {
            label->predecessor = new_active_label;
            update_queue.push_back(label);
            break;
         }
      }
   }

   if (collect_decision_points) this->data.updated_decision_nodes.clear();

   while (!update_queue.empty()) {
      label_t* const pred_label = update_queue.front();  update_queue.pop_front();
      pred_label->min_cost += weight_diff;
      pred_label->vars += vars_diff;

      if (this->data.shortest_path_tree[pred_label->edge_id]) {
         if (collect_decision_points && pred_label->sibling) {
            this->data.updated_decision_nodes.push_back(pred_label);
         }
         for (auto in_edge_it = entire(this->data.G.in_edges(pred_label->node));  !in_edge_it.at_end();  ++in_edge_it) {
            for (label_t* label = this->data.labels_on_node[in_edge_it.from_node()]; label; label = label->sibling) {
               if (label->predecessor == pred_label) {
                  update_queue.push_back(label);
                  break;
               }
            }
         }
      }
   }
}

template <typename Weight>
bool AllSolutions<Weight>::check_updated_decision_nodes() const
{
   for (const label_t* label : this->data.updated_decision_nodes) {
      if (!is_better_than_siblings(label))
         return false;
   }
   return true;
}

} }

#endif // POLYMAKE_POLYTROPES_GRAPH_ALL_SOLUTIONS_H

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
