#ifndef POLYMAKE_POLYTROPES_GRAPH_DIJKSTRA_SHORTEST_PATH_WITH_INTERVAL_WEIGHTS_H
#define POLYMAKE_POLYTROPES_GRAPH_DIJKSTRA_SHORTEST_PATH_WITH_INTERVAL_WEIGHTS_H

#include "polymake/Graph.h"
#include "polymake/SparseVector.h"
#include "polymake/graph/DijkstraShortestPathBase.h"
#include "polymake/graph/DijkstraShortestPathWithSiblings.h"
#include "polymake/GenericStruct.h"

#include <tuple>

namespace polymake {
namespace graph {

template <typename Weight>
class WeightInterval
   : public GenericStruct<WeightInterval<Weight>> {
public:
   DeclSTRUCT( DeclFIELD(var_index, Int)
               DeclTemplFIELD(min_weight, Weight)
               DeclTemplFIELD(max_weight, Weight) );

   WeightInterval() = default;
   
   WeightInterval(Int v, const Weight& min, const Weight& max)
   {
      var_index = v;
      min_weight = min;
      max_weight = max;
   }

   bool is_zero() const
   {
      return min_weight == 0 && var_index < 0;
   }
};

template <typename Weight>
class DijkstraShortestPathWithIntervalWeights
   : public DijkstraShortestPathWithSiblings {
   using base_layer = DijkstraShortestPathWithSiblings;
public:
   template <typename FullLabel = void>
   struct Label : public base_layer::Label<typename mprefer1st<FullLabel, Label<FullLabel>>::type> {
      using base_t = base_layer::Label<typename mprefer1st<FullLabel, Label<FullLabel>>::type>;

      template <typename CostArg, typename VarsArg>
      explicit Label(const Int node_arg, CostArg&& min_arg, CostArg&& max_arg, VarsArg&& vars_arg)
         : base_t(node_arg)
         , min_cost(std::forward<CostArg>(min_arg))
         , max_cost(std::forward<CostArg>(max_arg))
         , min_max_cost(max_cost)
         , vars(std::forward<VarsArg>(vars_arg)) {}

      const Weight& get_min_cost() const { return min_cost; }

      bool is_redundant_by_vars(const Label* other) const
      {
         // if this sibling label has all the same variables on its path towards the target node
         // as another label (and maybe some more variables), it can't overtake the latter
         return incl(other->vars, vars) <= 0;
      }

      bool is_worse_and_redundant_by_vars(const Label* other) const
      {
         return other->min_cost <= min_cost && is_redundant_by_vars(other);
      }

      bool is_redundant_by_min_cost(const Label* other)
      {
         if (other->min_max_cost < min_cost)
            return true;
         min_max_cost = other->min_max_cost;
         return false;
      }

      //! lower cost limit
      Weight min_cost;
      //! upper cost limit
      Weight max_cost;
      //! minimal upper cost limit over all siblings
      Weight min_max_cost;
      //! linear combination of variables on the path to target
      SparseVector<Int> vars;
   };

   using graph_t = Graph<Directed>;
   using label_comparator_t = LabelComparatorWithUpdates;

   // variable index, min weight, max weight
   using edge_weight_t = WeightInterval<Weight>;  // std::tuple<Int, Weight, Weight>;
   using edge_weights_map = EdgeMap<Directed, edge_weight_t>;

   template <typename Top>
   class Data : public base_layer::template Data<Top> {
      using base_t = typename base_layer::template Data<Top>;
   public:
      using typename base_t::label_t;
      using typename base_t::label_comparator_t;

      Data(const graph_t& G_arg, const edge_weights_map& edge_weights_arg, Int n_vars_arg)
         : base_t(G_arg)
         , edge_weights(edge_weights_arg)
         , n_vars(n_vars_arg) {}

      const edge_weights_map& edge_weights;
      const Int n_vars;
   };

   // similar to incl(Set, Set), taking multiplicities into account
   static Int incl(const SparseVector<Int>& v1, const SparseVector<Int>& v2)
   {
      const operations::cmp cmp{};
      auto it1 = entire(v1);
      auto it2 = entire(v2);
      Int result = cmp(v1.size(), v2.size());
      while (!it1.at_end() && !it2.at_end()) {
         switch (cmp(it1.index(), it2.index())) {
         case pm::cmp_eq: {
            switch (cmp(*it1, *it2)) {
            case pm::cmp_lt:
               if (result > 0) return 2;
               result = -1;
               break;
            case pm::cmp_gt:
               if (result < 0) return 2;
               result = 1;
               break;
            case pm::cmp_eq:
               break;
            }
            ++it1; ++it2;
            break;
         }
         case pm::cmp_lt:
            if (result < 0) return 2;
            result = 1;
            ++it1;
            break;
         case pm::cmp_gt:
            if (result > 0) return 2;
            result = -1;
            ++it2;
            break;
         }
      }
      if ((result < 0 && !it1.at_end()) || (result > 0 && !it2.at_end())) return 2;
      return result;
   }

   template <typename Top>
   class Algo : public base_layer::template Algo<Top> {
   public:
      using base_t = typename base_layer::template Algo<Top>;
      using typename base_t::label_t;
      using typename base_t::data_t;

      using base_layer::template Algo<Top>::Algo;

      label_t* construct_label(const Int node) const
      {
         return base_t::construct_label(node, zero_value<Weight>(), zero_value<Weight>(), SparseVector<Int>(this->data.n_vars));
      }

      label_t* construct_label(const label_t* const pred_label, Int cur_node, Int edge_id) const
      {
         const edge_weight_t& edge_weight = this->data.edge_weights[edge_id];
         label_t* new_label = base_t::construct_label(cur_node,
                                                      pred_label->min_cost + edge_weight.min_weight,
                                                      pred_label->min_max_cost + edge_weight.max_weight,
                                                      pred_label->vars);
         const Int var_index = edge_weight.var_index;
         if (var_index >= 0)
            ++(new_label->vars[var_index]);
         return new_label;
      }

      DijkstraLabelCmp compare_labels(label_t* const new_label, label_t* const old_label) const
      {
         if (old_label->min_max_cost <= new_label->min_cost) {
            // entire old interval below entire new interval, or both labels have equal costs with zero-width intervals
            return DijkstraLabelCmp::discard_new;
         }
         if (new_label->min_max_cost <= old_label->min_cost) {
            // entire new interval below entire old interval
            return DijkstraLabelCmp::discard_old_and_siblings;
         }

         const pm::cmp_value cmp_min_cost = operations::cmp()(new_label->min_cost, old_label->min_cost);
         const pm::cmp_value cmp_min_max_cost = operations::cmp()(new_label->min_max_cost, old_label->min_max_cost);

         if (cmp_min_cost == pm::cmp_lt) {
            // new label is better
            DijkstraLabelCmp result = DijkstraLabelCmp::replace_old;
            if (cmp_min_max_cost == pm::cmp_lt) {
               // and also improved the upper limit
               if (old_label->is_redundant_by_vars(new_label)) {
                  result = DijkstraLabelCmp::discard_old_keep_siblings;
               } else {
                  old_label->min_max_cost = new_label->min_max_cost;
               }
               this->remove_redundant_siblings(new_label, old_label, is_redundant_sibling());
            } else {
               if (cmp_min_max_cost == pm::cmp_gt)
                  new_label->min_max_cost = old_label->min_max_cost;
               this->remove_redundant_siblings(new_label, old_label,
                                               [](const label_t* main_label, label_t* sibling) {
                                                  return sibling->is_redundant_by_vars(main_label);
                                               });
            }
            return result;
         }

         if (cmp_min_max_cost == pm::cmp_lt) {
            old_label->min_max_cost = new_label->min_max_cost;
            this->remove_redundant_siblings(new_label, old_label, is_redundant_sibling());
            return DijkstraLabelCmp::keep_new_update_old;
         }

         if (new_label->is_redundant_by_vars(old_label))
            return DijkstraLabelCmp::discard_new;

         for (const label_t* old_sibling = old_label->sibling; old_sibling; old_sibling = old_sibling->sibling) {
            if (new_label->is_worse_and_redundant_by_vars(old_sibling))
               return DijkstraLabelCmp::discard_new;
         }

         if (cmp_min_max_cost == pm::cmp_gt)
            new_label->min_max_cost = old_label->min_max_cost;

         this->remove_redundant_siblings(new_label, old_label,
                                         [](const label_t* main_label, label_t* sibling) {
                                            return sibling->is_worse_and_redundant_by_vars(main_label);
                                         });
         return DijkstraLabelCmp::keep_new_as_sibling;
      }

      bool update_label(const label_t* const pred_label, label_t* const cur_label, Int edge_id) const
      {
         // an update can only change (reduce) the accumulated upper limit
         Weight new_max_cost = pred_label->min_max_cost + this->data.edge_weights[edge_id].max_weight;
         if (new_max_cost < cur_label->min_max_cost) {
            cur_label->min_max_cost = std::move(new_max_cost);
            return true;
         }
         return false;
      }

      auto is_redundant_sibling() const {
         return [](const label_t* main_label, label_t* sibling) {
            return sibling->is_redundant_by_min_cost(main_label) || sibling->is_redundant_by_vars(main_label);
         };
      }

      auto is_redundant_sibling_after_update() const
      {
         return [](const label_t* main_label, label_t* sibling) {
            return sibling->is_redundant_by_min_cost(main_label);
         };
      }
   };
};

} }

#endif // POLYMAKE_POLYTROPES_GRAPH_DIJKSTRA_SHORTEST_PATH_WITH_INTERVAL_WEIGHTS_H

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
