# Read a tntp file from https://github.com/bstabler/TransportationNetworks
# and create a graph with randomly created variable edges, controlled by a probability parameter.
# Probability for each edge independently, as in Erdös-Renyi.

# TNTP is tab delimited text files, with each row terminated by a semicolon. The files have the following format:
#
# First lines are metadata; each item has a description. An important one is the <FIRST THRU NODE>. In the some networks (like Sioux-Falls) it is equal to 1, indicating that traffic can move through all nodes, including zones. In other networks when traffic is not allow to go through zones, the zones are numbered 1 to n and the <FIRST THRU NODE> is set to n+1.
# Comment lines start with ‘~’.
# Network files – one line per link; links are directional, going from “init node” to “term node”.
#
# Link travel time = free flow time * ( 1 + B * (flow/capacity)^Power ).
# Link generalized cost = Link travel time + toll_factor * toll + distance_factor * distance
# The network files also contain a "speed" value for each link. In some cases the "speed" values are consistent with the free flow times, in other cases they represent posted speed limits, and in some cases there is no clear knowledge about their meaning. All of the results reported below are based only on free flow travel times as described by the functions above, and do not use the speed values.
# The standard order of the fields in the network files is:
# Init node, Term node, Capacity, Length, Free Flow Time, B, Power, Speed limit, Toll, Link Type

use application "graph";

print STDERR
    'call "$Network = read_tntp(filename,probability);" to produce graph from a network in tntp format', "\n",
    'call "add_more_vars($Network,n_more_vars);" to add a constant number of variable arcs', "\n",
    'call "$M = network2matrix($Network);"', "\n";

sub read_tntp($;$) {
  my ($tntp_file,$probability)=@_;
  open TNTP, $tntp_file or die "could not read $tntp_file\n";

  my $G = undef; # props::Graph<Directed> to be defined once number of nodes is known from meta data
  my $Weights;

  my %meta=();
  my $data=0;
  my $n_nodes = 0;
  my $n_zones = 0;
  my $n_links = 0;
  my $n_vars = 0;
  foreach my $line (<TNTP>) {
      chomp;
      if (!$data) {
	  my ($key, $number) = ( $line =~ /<(.*)>\s*(.*)/ );
	  $key = uc($key);
	  $key =~ s/ /_/g;
	  if ($key eq "END_OF_METADATA") {
	      $data=1;
	      $n_zones = defined($meta{NUMBER_OF_ZONES}) ? $meta{NUMBER_OF_ZONES} : 0;
	      $n_nodes = $meta{NUMBER_OF_NODES} - $n_zones;
	      $G = new props::Graph<Directed>($n_nodes);
	      $Weights = new EdgeMap<Directed, WeightInterval<Rational>>($G);
	  } else {
	      $meta{$key} = $number+0;
	  }
      } else {
	  next if ( ($line =~ /^\s*~/) || ($line =~ /^\s*$/) );
	  my ($init_node, $term_node, $capacity, $length, $free_flow_time, $bias, $power, $speed_limit, $toll, $type)
	      = ( $line =~ /\s*(\d+)\s+(\d+)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*[^\s]*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+)/ );
	  ++$n_links;
	  if ($init_node>$n_zones && $term_node>$n_zones) {
	      my $u = $init_node - $n_zones - 1; # ignore zones and ...
	      my $v = $term_node - $n_zones - 1; # numbering of nodes in tntp format starts with 1
	      if (rand() < $probability) { # variable edge with (low) probability
		  my $flow = rand() * $capacity; # random portion of the capacity
		  my $link_time = $free_flow_time * ( 1 + $bias * ($flow/$capacity)^$power ); # BPR formula
		  $Weights->edge($u,$v)=new WeightInterval<Rational>($n_vars, $free_flow_time, $link_time);
		  ++$n_vars;
	      } else { # constant edge
		  $Weights->edge($u,$v)=new WeightInterval<Rational>(-1, $free_flow_time, $free_flow_time);
	      }
	  }
      }
  }
  close TNTP or die "could not close $tntp_file\n";

  print STDERR "meta data found:";
  foreach (keys(%meta)) { print STDERR " $_=", $meta{$_}; }
  print STDERR "\n", "$n_links lines of data read";
  
  my $Network = new graph::Graph<Directed>(ADJACENCY=>$G, NODE_LABELS=>[($n_zones+1)..($n_zones+$n_nodes)]); # keep original node labels
  $Network->attach("N_ZONES", $n_zones);
  $Network->attach("N_LINKS", $n_links);
  $Network->attach("WEIGHTS", $Weights, "ADJACENCY");
  $Network->attach("PROBABILITY", $probability);
  $Network->attach("N_VARS", $n_vars);
  $Network->description = "network from $tntp_file, weights: mostly free flow time, some variable link times for random portion of capacity";
  ($Network->name) = ( $tntp_file =~ /([^.]*).*/ );
  $Network->name .= "+p$probability+v$n_vars";

  print STDERR ", the graph has ", $n_nodes, " proper nodes and ", $Network->N_EDGES, " arcs, $n_vars variables\n";
  return $Network;
}

sub add_more_vars($$) {
    my ($Network,$n_more_vars)=@_;
    my $n_edges = $Network->N_EDGES;

    die "graph only has $n_edges edges" if $n_more_vars>$n_edges;

    my $Weights=$Network->get_attachment("WEIGHTS");
    my $n_vars=$Network->get_attachment("N_VARS");
    my $k=$n_vars; # total number of vars
    
    my $random_indices=new Set();
    while ($random_indices->size()<$n_more_vars) {
	$random_indices += Int(rand()*$n_edges);
    }

    my $lidx=0; # link index
    my $infty = new Rational("inf");
    for (my $e=entire(edges($Network->ADJACENCY)); $e; ++$e, ++$lidx) {
	if ($random_indices->contains($lidx)) {
	    my ($u,$v)=($e->from_node,$e->to_node);
	    my ($i,$lower,$upper)=@{ $Weights->edge($u,$v) };
	    if ($i<0) { # new var
		$i = $k++;
		$Weights->edge($u,$v) = new WeightInterval<Rational>($i, $lower, 2*$lower);
	    }
	}
    }

    $Network->attach("WEIGHTS", $Weights, "ADJACENCY");
    $Network->attach("N_VARS", $k);
    $Network->name .= "-v$k";
    return $Network;
}

sub constant_polynomial($$) {
    my ($n_vars,$c)=@_;
    my $Tv = new Vector<TropicalNumber<Min>>([$c]);
    my $Tm = new Matrix<Int>([zero_vector<Int>($n_vars)]);
    return new Polynomial<TropicalNumber<Min>, Int>($Tv, $Tm);
}

sub interval2polynomial($$) {
    my ($n_vars,$w)=@_; # Int, WeightInterval<Rational>
    my ($i,$lower,$upper)=@{$w};
    if ($i<0) {
	return constant_polynomial($n_vars,$lower); # upper required to agree with lower here
    } else {
	my $Tv = new Vector<TropicalNumber<Min>>([$lower]); # ignore upper for now
	my $Tm = new Matrix<Int>([unit_vector<Int>($n_vars,$i)]);
	return new Polynomial<TropicalNumber<Min>, Int>($Tv, $Tm);
    }
}

sub network2matrix($) {
    my ($Network)=@_; # graph::Graph
    my $n_nodes=$Network->N_NODES;
    my $n_vars=$Network->get_attachment("N_VARS");
    my $inf = constant_polynomial($n_vars,"inf");
    my $zero = constant_polynomial($n_vars,0);
    my $M=new Matrix<Polynomial<TropicalNumber<Min>, Int>>($n_nodes,$n_nodes);
    for (my $i=0; $i<$n_nodes; ++$i) {
        for (my $j=0; $j<$n_nodes; ++$j) { $M->elem($i,$j)=$inf; }
    }
    for (my $i=0; $i<$n_nodes; ++$i) { $M->elem($i,$i)=$zero; }
    my $em=$Network->get_attachment("WEIGHTS");
    for (my $e=entire(edges($Network->ADJACENCY)); $e; ++$e) {
	my ($u,$v)=($e->from_node,$e->to_node);
	$M->elem($u,$v) = interval2polynomial($n_vars,$em->edge($u,$v));
    }
    return $M;
}

sub tex_matrix($) {
    my ($M)=@_;
    my $str="";
    for (my $i=0; $i<$M->rows(); ++$i) {
	for (my $j=0; $j<$M->cols(); ++$j) {
	    $str .= ' & ' unless $j==0;
	    my $elem=$M->elem($i,$j); $elem="$elem";
	    $elem =~ s/\s+//g;
	    my @terms = split /\+/, $elem;
	    my @new_terms = map { s/\*/\+/g; $_ } @terms; # multiplication is "+"
	    my $elem_str;
	    if ($#new_terms==0) {
		$elem_str = $new_terms[0] eq "inf" ? '\infty' : $new_terms[0];
	    } else {
		$elem_str = '\min( ' . (join ', ', @new_terms) . ' )'; # addition is min
	    }
	    $str .= $elem_str;
	}
	$str .= ' \\\\' unless $i==$M->rows()-1;
	$str .= "\n";
    }
    return $str;
}
