#include "polymake/graph/AllSolutions.h"
#include "polymake/Rational.h"
#include "polymake/client.h"

#include <ctime>
#include <ratio>
#include <chrono>

namespace polymake {
namespace graph {

void parameterized_shortest_path_trees(BigObject NET, Int target_node, Int verbose)
{
   const Graph<Directed>& G = NET.give("ADJACENCY");

   const AllSolutions<Rational>::edge_weights_map& edge_weights = NET.get_attachment("WEIGHTS");
   double probability = NET.get_attachment("PROBABILITY");
   Int n_vars = NET.get_attachment("N_VARS");

   using namespace std::chrono;
   high_resolution_clock::time_point t0 = high_resolution_clock::now();

   AllSolutions<Rational> sol(G, edge_weights, n_vars);
   sol.start(target_node); // target node, computes unparameterized shortest paths (for all minima)
   long sol_cnt=0; // number of combinatorially distinct generic solutions
   do {
      ++sol_cnt;
      auto dummy(sol.get_shortest_path_tree()); // triggers computation
      if (verbose>0) {
         cout << "solution # " << sol_cnt << "\nshortest paths:\n";
         for (auto edge_it = entire(edges(G)); !edge_it.at_end(); ++edge_it) {
            if (sol.get_shortest_path_tree()[*edge_it]) // EdgeMap<bool> to mark edges in shortest path tree
               cout << "  " << edge_it.from_node() << " -> " << edge_it.to_node() << "\n";
         }
         cout << "distances to target (constant + linear combination of variables):\n";
         for (Int n=0; n<G.nodes(); ++n) {
            cout << "  " << n << ": " << sol.get_cost(n) << "\n";
         }
         cout << "inequalities:\n" << sol.get_inequalities() << endl;
      }
   } while (sol.next());
   high_resolution_clock::time_point t1 = high_resolution_clock::now();
   duration<double> time_span = duration_cast<duration<double>>(t1 - t0);
   double perf = (double) sol_cnt/time_span.count();

   cout << "#nodes=" << G.nodes() << " #edges=" << G.edges() << " p=" << probability << " #vars=" << n_vars << " #solutions=" << sol_cnt;
   if (verbose>1) 
      cout  << " secs=" << time_span.count() << " perf=" << perf;
   cout << endl;
}

Function4perl(&parameterized_shortest_path_trees, "parameterized_shortest_path_trees(Graph, $ ; $=0)");

} }

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
